@@ -5,8 +5,10 @@
 from typing import Optional, Dict, List, Tuple, TYPE_CHECKING, NamedTuple, Callable
 from enum import Enum, auto
 
-from .util import bfh, bh2u
+from .util import bfh
 from .bitcoin import redeem_script_to_address, dust_threshold, construct_witness
+from .invoices import PR_PAID
+from . import descriptor
 from . import ecc
 from .lnutil import (make_commitment_output_to_remote_address, make_commitment_output_to_local_witness_script,
                      derive_privkey, derive_pubkey, derive_blinded_pubkey, derive_blinded_privkey,
@@ -25,12 +27,13 @@ if TYPE_CHECKING:
 
 
 _logger = get_logger(__name__)
+# note: better to use chan.logger instead, when applicable
 
 
 class SweepInfo(NamedTuple):
     name: str
     csv_delay: int
-    cltv_expiry: int
+    cltv_abs: int
     gen_tx: Callable[[], Optional[Transaction]]
 
 
@@ -51,8 +54,8 @@ def create_sweeptxs_for_watchtower(chan: 'Channel', ctx: Transaction, per_commit
     txs = []
     # to_local
     revocation_pubkey = ecc.ECPrivkey(other_revocation_privkey).get_public_key_bytes(compressed=True)
-    witness_script = bh2u(make_commitment_output_to_local_witness_script(
-        revocation_pubkey, to_self_delay, this_delayed_pubkey))
+    witness_script = make_commitment_output_to_local_witness_script(
+        revocation_pubkey, to_self_delay, this_delayed_pubkey).hex()
     to_local_address = redeem_script_to_address('p2wsh', witness_script)
     output_idxs = ctx.get_output_idxs_from_address(to_local_address)
     if output_idxs:
@@ -61,7 +64,7 @@ def create_sweeptxs_for_watchtower(chan: 'Channel', ctx: Transaction, per_commit
             sweep_address=sweep_address,
             ctx=ctx,
             output_idx=output_idx,
-            witness_script=witness_script,
+            witness_script=bfh(witness_script),
             privkey=other_revocation_privkey,
             is_revocation=True,
             config=chan.lnworker.config)
@@ -80,7 +83,6 @@ def create_sweeptxs_for_watchtower(chan: 'Channel', ctx: Transaction, per_commit
             htlc=htlc,
             ctx_output_idx=ctx_output_idx)
         return create_sweeptx_that_spends_htlctx_that_spends_htlc_in_ctx(
-            to_self_delay=0,
             htlc_tx=htlc_tx,
             htlctx_witness_script=htlc_tx_witness_script,
             sweep_address=sweep_address,
@@ -119,8 +121,8 @@ def create_sweeptx_for_their_revoked_ctx(
     txs = []
     # to_local
     revocation_pubkey = ecc.ECPrivkey(other_revocation_privkey).get_public_key_bytes(compressed=True)
-    witness_script = bh2u(make_commitment_output_to_local_witness_script(
-        revocation_pubkey, to_self_delay, this_delayed_pubkey))
+    witness_script = make_commitment_output_to_local_witness_script(
+        revocation_pubkey, to_self_delay, this_delayed_pubkey).hex()
     to_local_address = redeem_script_to_address('p2wsh', witness_script)
     output_idxs = ctx.get_output_idxs_from_address(to_local_address)
     if output_idxs:
@@ -129,7 +131,7 @@ def create_sweeptx_for_their_revoked_ctx(
             sweep_address=sweep_address,
             ctx=ctx,
             output_idx=output_idx,
-            witness_script=witness_script,
+            witness_script=bfh(witness_script),
             privkey=other_revocation_privkey,
             is_revocation=True,
             config=chan.lnworker.config)
@@ -159,24 +161,23 @@ def create_sweeptx_for_their_revoked_htlc(
     this_delayed_pubkey = derive_pubkey(this_conf.delayed_basepoint.pubkey, pcp)
     # same witness script as to_local
     revocation_pubkey = ecc.ECPrivkey(other_revocation_privkey).get_public_key_bytes(compressed=True)
-    witness_script = bh2u(make_commitment_output_to_local_witness_script(
-        revocation_pubkey, to_self_delay, this_delayed_pubkey))
+    witness_script = make_commitment_output_to_local_witness_script(
+        revocation_pubkey, to_self_delay, this_delayed_pubkey).hex()
     htlc_address = redeem_script_to_address('p2wsh', witness_script)
     # check that htlc_tx is a htlc
     if htlc_tx.outputs()[0].address != htlc_address:
         return
-    gen_tx = lambda: create_sweeptx_ctx_to_local(
+    gen_tx = lambda: create_sweeptx_that_spends_htlctx_that_spends_htlc_in_ctx(
         sweep_address=sweep_address,
-        ctx=htlc_tx,
-        output_idx=0,
-        witness_script=witness_script,
+        htlc_tx=htlc_tx,
+        htlctx_witness_script=bfh(witness_script),
         privkey=other_revocation_privkey,
         is_revocation=True,
         config=chan.lnworker.config)
     return SweepInfo(
         name='redeem_htlc2',
         csv_delay=0,
-        cltv_expiry=0,
+        cltv_abs=0,
         gen_tx=gen_tx)
 
 
@@ -202,19 +203,22 @@ def create_sweeptxs_for_our_ctx(
     our_htlc_privkey = derive_privkey(secret=int.from_bytes(our_conf.htlc_basepoint.privkey, 'big'),
                                        per_commitment_point=our_pcp).to_bytes(32, 'big')
     our_localdelayed_pubkey = our_localdelayed_privkey.get_public_key_bytes(compressed=True)
-    to_local_witness_script = bh2u(make_commitment_output_to_local_witness_script(
-        their_revocation_pubkey, to_self_delay, our_localdelayed_pubkey))
+    to_local_witness_script = make_commitment_output_to_local_witness_script(
+        their_revocation_pubkey, to_self_delay, our_localdelayed_pubkey).hex()
     to_local_address = redeem_script_to_address('p2wsh', to_local_witness_script)
-    # to remote address
-    bpk = their_conf.payment_basepoint.pubkey
-    their_payment_pubkey = bpk if chan.is_static_remotekey_enabled() else derive_pubkey(their_conf.payment_basepoint.pubkey, our_pcp)
-    to_remote_address = make_commitment_output_to_remote_address(their_payment_pubkey)
-    # test ctx
-    _logger.debug(f'testing our ctx: {to_local_address} {to_remote_address}')
-    if not ctx.get_output_idxs_from_address(to_local_address) \
-       and not ctx.get_output_idxs_from_address(to_remote_address):
+    to_remote_address = None
+    # test if this is our_ctx
+    found_to_local = bool(ctx.get_output_idxs_from_address(to_local_address))
+    if not chan.is_backup():
+        assert chan.is_static_remotekey_enabled()
+        their_payment_pubkey = their_conf.payment_basepoint.pubkey
+        to_remote_address = make_commitment_output_to_remote_address(their_payment_pubkey)
+        found_to_remote = bool(ctx.get_output_idxs_from_address(to_remote_address))
+    else:
+        found_to_remote = False
+    if not found_to_local and not found_to_remote:
         return
-    # we have to_local, to_remote.
+    chan.logger.debug(f'(lnsweep) found our ctx: {to_local_address} {to_remote_address}')
     # other outputs are htlcs
     # if they are spent, we need to generate the script
     # so, second-stage htlc sweep should not be returned here
@@ -227,7 +231,7 @@ def create_sweeptxs_for_our_ctx(
             sweep_address=sweep_address,
             ctx=ctx,
             output_idx=output_idx,
-            witness_script=to_local_witness_script,
+            witness_script=bfh(to_local_witness_script),
             privkey=our_localdelayed_privkey.get_secret_bytes(),
             is_revocation=False,
             to_self_delay=to_self_delay,
@@ -236,22 +240,21 @@ def create_sweeptxs_for_our_ctx(
         txs[prevout] = SweepInfo(
             name='our_ctx_to_local',
             csv_delay=to_self_delay,
-            cltv_expiry=0,
+            cltv_abs=0,
             gen_tx=sweep_tx)
     we_breached = ctn < chan.get_oldest_unrevoked_ctn(LOCAL)
     if we_breached:
-        _logger.info("we breached.")
+        chan.logger.info(f"(lnsweep) we breached. txid: {ctx.txid()}")
         # return only our_ctx_to_local, because we don't keep htlc_signatures for old states
         return txs
 
     # HTLCs
     def create_txns_for_htlc(
-            *, htlc: 'UpdateAddHtlc', htlc_direction: Direction,
-            ctx_output_idx: int, htlc_relative_idx: int):
-        if htlc_direction == RECEIVED:
-            preimage = chan.lnworker.get_preimage(htlc.payment_hash)
-        else:
-            preimage = None
+            *, htlc: 'UpdateAddHtlc',
+            htlc_direction: Direction,
+            ctx_output_idx: int,
+            htlc_relative_idx: int,
+            preimage: Optional[bytes]):
         htlctx_witness_script, htlc_tx = create_htlctx_that_spends_from_our_ctx(
             chan=chan,
             our_pcp=our_pcp,
@@ -274,12 +277,12 @@ def create_sweeptxs_for_our_ctx(
         txs[htlc_tx.inputs()[0].prevout.to_str()] = SweepInfo(
             name='first-stage-htlc',
             csv_delay=0,
-            cltv_expiry=htlc_tx.locktime,
+            cltv_abs=htlc_tx.locktime,
             gen_tx=lambda: htlc_tx)
         txs[htlc_tx.txid() + ':0'] = SweepInfo(
             name='second-stage-htlc',
             csv_delay=to_self_delay,
-            cltv_expiry=0,
+            cltv_abs=0,
             gen_tx=sweep_tx)
 
     # offered HTLCs, in our ctx --> "timeout"
@@ -291,11 +294,20 @@ def create_sweeptxs_for_our_ctx(
         subject=LOCAL,
         ctn=ctn)
     for (direction, htlc), (ctx_output_idx, htlc_relative_idx) in htlc_to_ctx_output_idx_map.items():
+        if direction == RECEIVED:
+            if chan.lnworker.get_payment_status(htlc.payment_hash) == PR_PAID:
+                preimage = chan.lnworker.get_preimage(htlc.payment_hash)
+            else:
+                # do not redeem this, it might publish the preimage of an incomplete MPP
+                continue
+        else:
+            preimage = None
         create_txns_for_htlc(
             htlc=htlc,
             htlc_direction=direction,
             ctx_output_idx=ctx_output_idx,
-            htlc_relative_idx=htlc_relative_idx)
+            htlc_relative_idx=htlc_relative_idx,
+            preimage=preimage)
     return txs
 
 
@@ -319,7 +331,7 @@ def analyze_ctx(chan: 'Channel', ctx: Transaction):
             return
         their_pcp = ecc.ECPrivkey(per_commitment_secret).get_public_key_bytes(compressed=True)
         is_revocation = True
-        #_logger.info(f'tx for revoked: {list(txs.keys())}')
+        #chan.logger.debug(f'(lnsweep) tx for revoked: {list(txs.keys())}')
     elif chan.get_data_loss_protect_remote_pcp(ctn):
         their_pcp = chan.get_data_loss_protect_remote_pcp(ctn)
         is_revocation = False
@@ -345,18 +357,22 @@ def create_sweeptxs_for_their_ctx(
     # to_local and to_remote addresses
     our_revocation_pubkey = derive_blinded_pubkey(our_conf.revocation_basepoint.pubkey, their_pcp)
     their_delayed_pubkey = derive_pubkey(their_conf.delayed_basepoint.pubkey, their_pcp)
-    witness_script = bh2u(make_commitment_output_to_local_witness_script(
-        our_revocation_pubkey, our_conf.to_self_delay, their_delayed_pubkey))
+    witness_script = make_commitment_output_to_local_witness_script(
+        our_revocation_pubkey, our_conf.to_self_delay, their_delayed_pubkey).hex()
     to_local_address = redeem_script_to_address('p2wsh', witness_script)
-    # to remote address
-    bpk = our_conf.payment_basepoint.pubkey
-    our_payment_pubkey = bpk if chan.is_static_remotekey_enabled() else derive_pubkey(bpk, their_pcp)
-    to_remote_address = make_commitment_output_to_remote_address(our_payment_pubkey)
+    to_remote_address = None
     # test if this is their ctx
-    _logger.debug(f'testing their ctx: {to_local_address} {to_remote_address}')
-    if not ctx.get_output_idxs_from_address(to_local_address) \
-       and not ctx.get_output_idxs_from_address(to_remote_address):
+    found_to_local = bool(ctx.get_output_idxs_from_address(to_local_address))
+    if not chan.is_backup():
+        assert chan.is_static_remotekey_enabled()
+        our_payment_pubkey = our_conf.payment_basepoint.pubkey
+        to_remote_address = make_commitment_output_to_remote_address(our_payment_pubkey)
+        found_to_remote = bool(ctx.get_output_idxs_from_address(to_remote_address))
+    else:
+        found_to_remote = False
+    if not found_to_local and not found_to_remote:
         return
+    chan.logger.debug(f'(lnsweep) found their ctx: {to_local_address} {to_remote_address}')
     if is_revocation:
         our_revocation_privkey = derive_blinded_privkey(our_conf.revocation_basepoint.privkey, per_commitment_secret)
         gen_tx = create_sweeptx_for_their_revoked_ctx(chan, ctx, per_commitment_secret, chan.sweep_address)
@@ -365,51 +381,28 @@ def create_sweeptxs_for_their_ctx(
             txs[tx.inputs()[0].prevout.to_str()] = SweepInfo(
                 name='to_local_for_revoked_ctx',
                 csv_delay=0,
-                cltv_expiry=0,
+                cltv_abs=0,
                 gen_tx=gen_tx)
     # prep
     our_htlc_privkey = derive_privkey(secret=int.from_bytes(our_conf.htlc_basepoint.privkey, 'big'), per_commitment_point=their_pcp)
     our_htlc_privkey = ecc.ECPrivkey.from_secret_scalar(our_htlc_privkey)
     their_htlc_pubkey = derive_pubkey(their_conf.htlc_basepoint.pubkey, their_pcp)
     # to_local is handled by lnwatcher
-    # to_remote
-    if not chan.is_static_remotekey_enabled():
-        our_payment_bp_privkey = ecc.ECPrivkey(our_conf.payment_basepoint.privkey)
-        our_payment_privkey = derive_privkey(our_payment_bp_privkey.secret_scalar, their_pcp)
-        our_payment_privkey = ecc.ECPrivkey.from_secret_scalar(our_payment_privkey)
-        assert our_payment_pubkey == our_payment_privkey.get_public_key_bytes(compressed=True)
-        output_idxs = ctx.get_output_idxs_from_address(to_remote_address)
-        if output_idxs:
-            output_idx = output_idxs.pop()
-            prevout = ctx.txid() + ':%d'%output_idx
-            sweep_tx = lambda: create_sweeptx_their_ctx_to_remote(
-                sweep_address=sweep_address,
-                ctx=ctx,
-                output_idx=output_idx,
-                our_payment_privkey=our_payment_privkey,
-                config=chan.lnworker.config)
-            txs[prevout] = SweepInfo(
-                name='their_ctx_to_remote',
-                csv_delay=0,
-                cltv_expiry=0,
-                gen_tx=sweep_tx)
     # HTLCs
     def create_sweeptx_for_htlc(
-            htlc: 'UpdateAddHtlc', is_received_htlc: bool,
-            ctx_output_idx: int) -> None:
-        if not is_received_htlc and not is_revocation:
-            preimage = chan.lnworker.get_preimage(htlc.payment_hash)
-        else:
-            preimage = None
+            *, htlc: 'UpdateAddHtlc',
+            is_received_htlc: bool,
+            ctx_output_idx: int,
+            preimage: Optional[bytes]) -> None:
         htlc_output_witness_script = make_htlc_output_witness_script(
             is_received_htlc=is_received_htlc,
             remote_revocation_pubkey=our_revocation_pubkey,
             remote_htlc_pubkey=our_htlc_privkey.get_public_key_bytes(compressed=True),
             local_htlc_pubkey=their_htlc_pubkey,
             payment_hash=htlc.payment_hash,
-            cltv_expiry=htlc.cltv_expiry)
+            cltv_abs=htlc.cltv_abs)
 
-        cltv_expiry = htlc.cltv_expiry if is_received_htlc and not is_revocation else 0
+        cltv_abs = htlc.cltv_abs if is_received_htlc and not is_revocation else 0
         prevout = ctx.txid() + ':%d'%ctx_output_idx
         sweep_tx = lambda: create_sweeptx_their_ctx_htlc(
             ctx=ctx,
@@ -419,12 +412,12 @@ def create_sweeptxs_for_their_ctx(
             output_idx=ctx_output_idx,
             privkey=our_revocation_privkey if is_revocation else our_htlc_privkey.get_secret_bytes(),
             is_revocation=is_revocation,
-            cltv_expiry=cltv_expiry,
+            cltv_abs=cltv_abs,
             config=chan.lnworker.config)
         txs[prevout] = SweepInfo(
             name=f'their_ctx_htlc_{ctx_output_idx}',
             csv_delay=0,
-            cltv_expiry=cltv_expiry,
+            cltv_abs=cltv_abs,
             gen_tx=sweep_tx)
     # received HTLCs, in their ctx --> "timeout"
     # offered HTLCs, in their ctx --> "success"
@@ -435,18 +428,32 @@ def create_sweeptxs_for_their_ctx(
         subject=REMOTE,
         ctn=ctn)
     for (direction, htlc), (ctx_output_idx, htlc_relative_idx) in htlc_to_ctx_output_idx_map.items():
+        is_received_htlc = direction == RECEIVED
+        if not is_received_htlc and not is_revocation:
+            if chan.lnworker.get_payment_status(htlc.payment_hash) == PR_PAID:
+                preimage = chan.lnworker.get_preimage(htlc.payment_hash)
+            else:
+                # do not redeem this, it might publish the preimage of an incomplete MPP
+                continue
+        else:
+            preimage = None
         create_sweeptx_for_htlc(
             htlc=htlc,
-            is_received_htlc=direction == RECEIVED,
-            ctx_output_idx=ctx_output_idx)
+            is_received_htlc=is_received_htlc,
+            ctx_output_idx=ctx_output_idx,
+            preimage=preimage)
     return txs
 
 
 def create_htlctx_that_spends_from_our_ctx(
-        chan: 'Channel', our_pcp: bytes,
-        ctx: Transaction, htlc: 'UpdateAddHtlc',
-        local_htlc_privkey: bytes, preimage: Optional[bytes],
-        htlc_direction: Direction, htlc_relative_idx: int,
+        chan: 'Channel',
+        our_pcp: bytes,
+        ctx: Transaction,
+        htlc: 'UpdateAddHtlc',
+        local_htlc_privkey: bytes,
+        preimage: Optional[bytes],
+        htlc_direction: Direction,
+        htlc_relative_idx: int,
         ctx_output_idx: int) -> Tuple[bytes, Transaction]:
     assert (htlc_direction == RECEIVED) == bool(preimage), 'preimage is required iff htlc is received'
     preimage = preimage or b''
@@ -460,7 +467,7 @@ def create_htlctx_that_spends_from_our_ctx(
         commit=ctx,
         htlc=htlc,
         ctx_output_idx=ctx_output_idx,
-        name=f'our_ctx_{ctx_output_idx}_htlc_tx_{bh2u(htlc.payment_hash)}')
+        name=f'our_ctx_{ctx_output_idx}_htlc_tx_{htlc.payment_hash.hex()}')
     remote_htlc_sig = chan.get_remote_htlc_sig_for_htlc(htlc_relative_idx=htlc_relative_idx)
     local_htlc_sig = bfh(htlc_tx.sign_txin(0, local_htlc_privkey))
     txin = htlc_tx.inputs()[0]
@@ -473,8 +480,8 @@ def create_sweeptx_their_ctx_htlc(
         ctx: Transaction, witness_script: bytes, sweep_address: str,
         preimage: Optional[bytes], output_idx: int,
         privkey: bytes, is_revocation: bool,
-        cltv_expiry: int, config: SimpleConfig) -> Optional[PartialTransaction]:
-    assert type(cltv_expiry) is int
+        cltv_abs: int, config: SimpleConfig) -> Optional[PartialTransaction]:
+    assert type(cltv_abs) is int
     preimage = preimage or b''  # preimage is required iff (not is_revocation and htlc is offered)
     val = ctx.outputs()[output_idx].value
     prevout = TxOutpoint(txid=bfh(ctx.txid()), out_idx=output_idx)
@@ -488,7 +495,7 @@ def create_sweeptx_their_ctx_htlc(
     outvalue = val - fee
     if outvalue <= dust_threshold(): return None
     sweep_outputs = [PartialTxOutput.from_address_and_value(sweep_address, outvalue)]
-    tx = PartialTransaction.from_io(sweep_inputs, sweep_outputs, version=2, locktime=cltv_expiry)
+    tx = PartialTransaction.from_io(sweep_inputs, sweep_outputs, version=2, locktime=cltv_abs)
     sig = bfh(tx.sign_txin(0, privkey))
     if not is_revocation:
         witness = construct_witness([sig, preimage, witness_script])
@@ -509,9 +516,8 @@ def create_sweeptx_their_ctx_to_remote(
     prevout = TxOutpoint(txid=bfh(ctx.txid()), out_idx=output_idx)
     txin = PartialTxInput(prevout=prevout)
     txin._trusted_value_sats = val
-    txin.script_type = 'p2wpkh'
-    txin.pubkeys = [bfh(our_payment_pubkey)]
-    txin.num_sig = 1
+    desc = descriptor.get_singlesig_descriptor_from_legacy_leaf(pubkey=our_payment_pubkey, script_type='p2wpkh')
+    txin.script_descriptor = desc
     sweep_inputs = [txin]
     tx_size_bytes = 110  # approx size of p2wpkh->p2wpkh
     fee = config.estimate_fee(tx_size_bytes, allow_fallback_to_static_rates=True)
@@ -527,9 +533,9 @@ def create_sweeptx_their_ctx_to_remote(
 
 
 def create_sweeptx_ctx_to_local(
-        *, sweep_address: str, ctx: Transaction, output_idx: int, witness_script: str,
+        *, sweep_address: str, ctx: Transaction, output_idx: int, witness_script: bytes,
         privkey: bytes, is_revocation: bool, config: SimpleConfig,
-        to_self_delay: int=None) -> Optional[PartialTransaction]:
+        to_self_delay: int = None) -> Optional[PartialTransaction]:
     """Create a txn that sweeps the 'to_local' output of a commitment
     transaction into our wallet.
 
@@ -541,7 +547,7 @@ def create_sweeptx_ctx_to_local(
     txin = PartialTxInput(prevout=prevout)
     txin._trusted_value_sats = val
     txin.script_sig = b''
-    txin.witness_script = bfh(witness_script)
+    txin.witness_script = witness_script
     sweep_inputs = [txin]
     if not is_revocation:
         assert isinstance(to_self_delay, int)
@@ -561,26 +567,20 @@ def create_sweeptx_ctx_to_local(
 
 def create_sweeptx_that_spends_htlctx_that_spends_htlc_in_ctx(
         *, htlc_tx: Transaction, htlctx_witness_script: bytes, sweep_address: str,
-        privkey: bytes, is_revocation: bool, to_self_delay: int,
+        privkey: bytes, is_revocation: bool, to_self_delay: int = None,
         config: SimpleConfig) -> Optional[PartialTransaction]:
-    val = htlc_tx.outputs()[0].value
-    prevout = TxOutpoint(txid=bfh(htlc_tx.txid()), out_idx=0)
-    txin = PartialTxInput(prevout=prevout)
-    txin._trusted_value_sats = val
-    txin.script_sig = b''
-    txin.witness_script = htlctx_witness_script
-    sweep_inputs = [txin]
-    if not is_revocation:
-        assert isinstance(to_self_delay, int)
-        sweep_inputs[0].nsequence = to_self_delay
-    tx_size_bytes = 200  # TODO
-    fee = config.estimate_fee(tx_size_bytes, allow_fallback_to_static_rates=True)
-    outvalue = val - fee
-    if outvalue <= dust_threshold(): return None
-    sweep_outputs = [PartialTxOutput.from_address_and_value(sweep_address, outvalue)]
-    tx = PartialTransaction.from_io(sweep_inputs, sweep_outputs, version=2)
-    sig = bfh(tx.sign_txin(0, privkey))
-    witness = construct_witness([sig, int(is_revocation), htlctx_witness_script])
-    tx.inputs()[0].witness = bfh(witness)
-    assert tx.is_complete()
-    return tx
+    """Create a txn that sweeps the output of a second stage htlc tx
+    (i.e. sweeps from an HTLC-Timeout or an HTLC-Success tx).
+    """
+    # note: this is the same as sweeping the to_local output of the ctx,
+    #       as these are the same script (address-reuse).
+    return create_sweeptx_ctx_to_local(
+        sweep_address=sweep_address,
+        ctx=htlc_tx,
+        output_idx=0,
+        witness_script=htlctx_witness_script,
+        privkey=privkey,
+        is_revocation=is_revocation,
+        to_self_delay=to_self_delay,
+        config=config,
+    )
